/// <reference types="node" />
import { Readable, Transform } from 'streamx';
import { EventEmitter } from 'events';
import { Schema, Store } from '@arsonar/common';
import { Files } from './files.js';
import type { Workspace } from './workspace.js';
import type { Logger, Record, RecordVersion, WireRecordVersion, RecordVersionForm, TypeSpecInput, SchemaSpec } from '@arsonar/common';
import { FetchOpts } from './fetch.js';
export declare type Recordlike = Record | RecordVersion | WireRecordVersion | RecordVersionForm;
export declare type GetRequest = {
    key: string;
    seq: string;
} | {
    type?: string;
    id?: string;
} | {
    lseq: number;
};
export interface GetOpts {
    sync?: boolean;
}
export interface CollectionInfo {
    id: string;
    name: string;
    key: string;
    discoveryKey: string;
    rootKey: string;
    localKey: string;
    length: number;
    feeds: FeedInfo[];
    peers: any;
    schema: SchemaSpec;
}
export interface FeedInfo {
    key: string;
    discoveryKey: string;
    length: number;
    byteLength: number;
}
export declare class Collection extends EventEmitter {
    readonly opened: boolean;
    readonly endpoint: string;
    readonly workspace: Workspace;
    readonly files: Files;
    readonly log: Logger;
    readonly schema: Schema;
    readonly store: Store;
    private _info;
    private readonly _nameOrKey;
    private _eventStream?;
    private _eventSource?;
    private _length;
    private readonly _cacheid?;
    private _liveUpdates;
    static uuid(): string;
    /**
       * Remote collection
       *
       * @constructor
       * @param workspace - Remote workspace
       * @param nameOrKey - Name or key of the collection
       */
    constructor(workspace: Workspace, nameOrKey: string, info: CollectionInfo);
    static open(workspace: Workspace, nameOrKey: string): Promise<Collection>;
    get name(): string;
    get key(): string | null;
    get localKey(): string | null;
    get info(): CollectionInfo | null;
    get id(): string | null;
    get length(): number;
    /**
     * Update collection info and schema from endpoint.
     */
    updateInfo(): Promise<void>;
    /**
       * Close the collection.
       *
       * Properly closes open HTTP requests.
       * @async
       */
    close(): Promise<void>;
    /**
       * Put a new feed to the collection.
       *
       * @async
       * @param key - The hex-encoded key of the feed to add.
       * @param [info] - Optional information about the feed.
       *                          TODO: Document
       */
    putFeed(key: string, info?: {}): Promise<any>;
    /**
     * Put a new feed to the collection.
     *
     * @deprecated replaced with putFeed.
     */
    addFeed(key: string, info?: {}): Promise<any>;
    /**
       * Query the collection.
       *
       * Returns an array of matching records.
       *
       * Records may have a meta property that includes query-specific metadata (e.g. the score for search queries).
       *
       * @async
       * @param {string} name - The name of a supported query.
       *                        Supported queries that ship with @arsonar/core are:
       *                        records, search, relations, history and indexes.
       * @param {object} args - The arguments for the query. Depends on the query being used.
       *                        For records: `{ schema, name, id }`
       *                        For history: `{ from: timestamp, to: timestamp }`
       *                        For search: Either a "string" for a simple full-text search, or a
       *                           tantivy query object.
       *                        For indexes: `{ schema, prop, value, from, to, reverse, limit }`
       *                           (to be documented)
       *                        For relations: `{ subject, object, predicate }`
       *                           where subject and object are ids and predicate is `type#field`
       *
       * @param {object} [opts] - Optional options
       * @param {boolean} [opts.sync=false] Wait for all pending indexing operations to be finished.
       *
       * @return {Promise<Array<Record>>} A promise that resolves to an array of record objects.
       */
    query(name: string, args: any, opts?: any): Promise<Record[]>;
    /**
       * Put a new record into the collection.
       *
       * @async
       * @param {object} record - The record.
       * @param {string} record.schema - The schema of the record.
       * @param {string} [record.id] - The entity id of the record. If empoty an id will be created.
       * @param {object} record.value - Value of the record.
       * @throws Throws if the record is invalid.
       * @return {Promise<object>} An object with an `{ id }` property.
       */
    put(record: RecordVersion | WireRecordVersion | RecordVersionForm): Promise<Record>;
    /**
       * Get records by their semantic address (type and id) or by their storage address (key and seq).
       *
       * @async
       * @param {object} req - The get request. Either `{ type, id }` or `{ key, seq }`.
       * @param {object} [opts] - Optional options.
       * @param {boolean} [opts.sync=false] Wait for all pending indexing operations to be finished.
       * @return {Promise<Array<object>>} A promise that resolves to an array of record objects.
       */
    get(req: GetRequest, opts?: GetOpts): Promise<Record[]>;
    /**
       * Get a specific version of a record.
       *
       * @async
       * @param {string} address - The block address of the record version `feedkey@seq`
       *    where `feedkey` is the hex-encoded public key of a feed and `seq` is a sequence number (uint).
       */
    getVersion(address: string): Promise<Record>;
    /**
       * Deletes a record.
       *
       * @async
       * @param {object} record - The record to delete. Has to have `{ id, type }` properties set.
       * @return {Promise<object>} - An object with `{ id, type }` properties of the deleted record.
       */
    del(record: Record | RecordVersion | WireRecordVersion): Promise<void>;
    /**
       * Add a new type to the collection.
       *
       * @async
       * @param {object} schema - A schema object.
       * @throws Throws if the schema object is invalid or cannot be saved.
       * @return {Promise<object>} A promise that resolves to the saved schema object.
       */
    putType(spec: TypeSpecInput): Promise<any>;
    /**
       * Create a writable stream to put records into the collection.
       *
       * Example:
       * ```javascript
       * const batchStream = collection.createBatchStream()
       * batch.write(record)
       * batch.close()
       * ```
       *
       * @return {Writable<Record>} A writable stream
       */
    createBatchStream(): Transform<Recordlike, string, string> & {
        finished: Promise<void>;
    };
    /**
       * Wait for all pending indexing operations to be finished.
       *
       * @async
       * @return {Promise<void>}
       */
    sync(): Promise<any>;
    /**
       * Subscribe to events on this collection.
       *
       * Returns a readable stream that emits event objects.
       * They look like this:
       * `{ event: string, data: object }`
       *
       * Events are:
       *
       * * `update`: with data `{ lseq }`
       * * `feed`: with data `{ key }`
       * * `schema-update`
       *
       * @return {Readable<object>}
       */
    createEventStream(): Readable<any>;
    private _initEventSource;
    /**
       * Pull live updates from the server as they happen.
       *
       * After calling this method once, all new records and record versions
       * are pulled from the server once available. The `update` event
       * is emitted when new records are about to arrive.
       */
    pullLiveUpdates(): void;
    /**
       * Subscribe to this collection.
       *
       * This will fetch all records from the first to the last and then waits for new records.
       * Currently only intended for usage in bots (not in short-running Browser clients).
       *
       * @todo: Prefix client ID to subscription name.
       * @todo: Allow to subscribe from now instead of from the beginning.
       *
       * @param {string} Subscription name. Has to be unique per running Sonar instance
       * @param {function} Async callback function that will be called for each incoming record
       */
    subscribe(name: string, onRecord: (record: Record) => Promise<void>): Promise<boolean>;
    private _pullSubscription;
    private _ackSubscription;
    /**
       * Reindex the secondary indexes (views) in this collection.
       *
       * Use with care, this can be expensive.
       *
       * @param {Array<string>} Optional array of view names to reindex.
       *  If unset all views will be reindexed.
       */
    reindex(views?: string[]): Promise<any>;
    fetch(path: string, opts?: FetchOpts): Promise<any>;
}
//# sourceMappingURL=collection.d.ts.map